/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.outgoing;

import static org.ow2.petals.binding.soap.Constants.ServiceUnit.ADD_ROOT;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.BASIC_AUTH_PASSWORD;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.BASIC_AUTH_USERNAME;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.CHUNKED_MODE;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.CLEANUP_TRANSPORT;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.PROXY_DOMAIN;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.PROXY_HOST;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.PROXY_PASSWORD;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.PROXY_PORT;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.PROXY_USER;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.SOAP_ACTION;
import static org.ow2.petals.binding.soap.Constants.ServiceUnit.SOAP_VERSION;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.apache.axiom.soap.SOAP11Constants;
import org.apache.axiom.soap.SOAP12Constants;
import org.apache.axis2.client.Options;
import org.apache.axis2.transport.http.HttpTransportProperties;
import org.apache.commons.httpclient.auth.AuthPolicy;
import org.ow2.petals.binding.soap.SoapComponentContext;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;

/**
 * Abstract class for dispatchers. Dispatchers are used to send JBI message to
 * external Web Services.
 * 
 * @author chamerling - eBM WebSourcing
 * 
 */
public abstract class AbstractExternalServiceCaller implements ExternalServiceCaller {

    public static final String SOAP_VERSION_11 = "1.1";

    public static final String SOAP_VERSION_12 = "1.2";

    /**
     * The soap component context
     */
    protected SoapComponentContext soapContext;

    /**
     * The logger
     */
    protected Logger logger;

    /**
     * Creates a new instance of {@link AbstractExternalServiceCaller}
     * 
     * @param soapContext
     * @param logger
     */
    public AbstractExternalServiceCaller(final SoapComponentContext soapContext, final Logger logger) {
        this.soapContext = soapContext;
        this.logger = logger;
    }

    /**
     * 
     * @param extensions
     * @return
     */
    protected final static boolean retrieveAddRoot(final ConfigurationExtensions extensions) {
        final String result = extensions.get(ADD_ROOT);
        return Boolean.parseBoolean(result);
    }

    /**
     * Retrieve the default SOAP action from the extensions.
     * 
     * @param extensions
     * @return
     */
    protected final static QName retrieveDefaultSOAPAction(final ConfigurationExtensions extensions) {
        QName soapAction = null;
        final String tmp = extensions.get(SOAP_ACTION);

        if (tmp != null) {
            soapAction = QName.valueOf(tmp);
        }

        return soapAction;
    }

    /**
     * Retrieve the proxy settings from the extensions. The proxy-host value is
     * required. The other ones will be set to default values by Axis if they
     * have not been setted by the SU.
     * 
     * 
     * @param extensions
     *            must be not null. Contains <em>proxy-*</em> optional tags.
     * 
     * @return the proxy settings if they are present in the extensions, null
     *         otherwise.
     * @see org.apache.axis2.transport.http.HttpTransportProperties.ProxyProperties
     * 
     */
    protected final static HttpTransportProperties.ProxyProperties retrieveProxySettings(
            final ConfigurationExtensions extensions) {

        HttpTransportProperties.ProxyProperties proxyProperties = null;

        if (extensions.get(PROXY_HOST) != null) {
            // proxy host is required, if it is not set the proxy mode is not
            // activated

            proxyProperties = new HttpTransportProperties.ProxyProperties();

            final String domain = extensions.get(PROXY_DOMAIN);
            final String password = extensions.get(PROXY_PASSWORD);
            final String proxyHost = extensions.get(PROXY_HOST);

            int proxyPort = -1;
            final String tmp = extensions.get(PROXY_PORT);
            if (tmp != null) {
                try {
                    proxyPort = Integer.parseInt(tmp);
                } catch (final NumberFormatException e) {
                }
            }

            final String proxyUserName = extensions.get(PROXY_USER);

            proxyProperties.setDomain(domain);
            proxyProperties.setPassWord(password);
            proxyProperties.setProxyName(proxyHost);
            proxyProperties.setProxyPort(proxyPort);
            proxyProperties.setUserName(proxyUserName);
        }

        return proxyProperties;
    }

    /**
     * Retrieve, in the jbi.xml descriptor, the SOAP version to use for
     * SoapEnvelopeNamespaceURI Search a <em>soap-version</em> tag. <br>
     * Legal values are <b>1.1</b> and <b>1.2</b>. 1.1 is used if no tag is
     * found.
     * 
     * @param extensions
     *            must be non null. Contains a <em>soap-version</em> optional
     *            tag.
     * @return the Axis2 soapEnvelopeNamespaceURI constant.
     * @see org.apache.axiom.soap.SOAP11Constants.SOAP_ENVELOPE_NAMESPACE_URI
     * @see org.apache.axiom.soap.SOAP12Constants.SOAP_ENVELOPE_NAMESPACE_URI
     */
    protected final static String retrieveSOAPEnvelopeNamespaceURI(final ConfigurationExtensions extensions) {
        String soapEnvelopeNamespaceURI = SOAP11Constants.SOAP_ENVELOPE_NAMESPACE_URI;
        final String soapVersion = extensions.get(SOAP_VERSION);

        if (SOAP_VERSION_12.equals(soapVersion)) {
            soapEnvelopeNamespaceURI = SOAP12Constants.SOAP_ENVELOPE_NAMESPACE_URI;
        }
        return soapEnvelopeNamespaceURI;
    }

    /**
     * Get the cleanup-transport property value (true/false).
     * 
     * @see bug #308219
     * @param extensions
     * @return
     */
    protected final static boolean retrieveCleanupTransport(final ConfigurationExtensions extensions) {
        final String clean = extensions.get(CLEANUP_TRANSPORT);

        return Boolean.valueOf(clean);
    }

    /**
     * Retrieve the chunked mode
     * 
     * @param extensions
     * @return
     */
    protected final static String retrieveChunkedMode(final ConfigurationExtensions extensions) {
        return extensions.get(CHUNKED_MODE);
    }

    /**
     * Getting the basic authentication parameters for the outgoing SOAP
     * request.
     * 
     * @param extensions
     * @param options
     * @return
     */
    @SuppressWarnings("unchecked")
    protected final static HttpTransportProperties.Authenticator retrieveBasicAuthentication(
            final ConfigurationExtensions extensions, final Options options) {
        HttpTransportProperties.Authenticator basicAuthentication = null;
        final String username = extensions.get(BASIC_AUTH_USERNAME);
        String password = extensions.get(BASIC_AUTH_PASSWORD);

        if (username != null) {
            if (password == null) {
                password = "";
            }

            basicAuthentication = new HttpTransportProperties.Authenticator();
            final List schemes = new ArrayList();
            schemes.add(AuthPolicy.BASIC);
            basicAuthentication.setAuthSchemes(schemes);
            basicAuthentication.setUsername(username);
            basicAuthentication.setPassword(password);
            basicAuthentication.setPreemptiveAuthentication(true);
            options.setProperty(org.apache.axis2.transport.http.HTTPConstants.AUTHENTICATE,
                    basicAuthentication);
        }

        return basicAuthentication;
    }
}
