/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.outgoing;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMText;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axiom.soap.impl.dom.soap11.SOAP11Factory;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.ow2.petals.binding.soap.SoapComponentContext;
import org.ow2.petals.binding.soap.listener.SUPropertiesHelper;
import org.ow2.petals.binding.soap.util.AxiomUtils;
import org.ow2.petals.binding.soap.util.Marshaller;
import org.ow2.petals.binding.soap.util.StaxUtils;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.performance.PerformanceNotification;
import org.w3c.dom.DocumentFragment;

import static javax.jbi.messaging.NormalizedMessageProperties.PROTOCOL_HEADERS;

/**
 * An external web service dispatcher. This dispatcher send the JBI message to
 * an external web service. The service URL is specified in the PEtALS
 * extensions.
 * 
 * @author chamerling - eBM WebSourcing
 * 
 */
public class SOAPCaller extends AbstractExternalServiceCaller {

    /**
     * The JBI component instance
     */
    private AbstractComponent component = null;

    /**
     * Creates a new instance of {@link SOAPCaller}
     * 
     * @param soapContext
     * @param logger
     */
    public SOAPCaller(final SoapComponentContext soapContext, final Logger logger,
            final AbstractComponent component) {
        super(soapContext, logger);
        this.component = component;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.binding.soap.listener.outgoing.JBIListenerDispatcher #
     *      process(org.objectweb.petals.component.common.util.MessageExchangeWrapper ,
     *      org.objectweb.petals.common.descriptor.Extensions)
     */
    public void call(final Exchange exchange, final ConfigurationExtensions cdkExtensions) {
        final String address = SUPropertiesHelper.getAddress(cdkExtensions);

        this.logger.fine("Calling external web service : " + address);

        // Get the incoming Normalized message
        final NormalizedMessage in = exchange.getInMessage();
        if (in == null) {
            exchange.setError(new JBIException(
                    "Message exchange must handle an In normalized message"));
        } else {

            // Get the outgoing Normalized message
            final NormalizedMessage out = exchange.getOutMessage();

            // Get operation
            final QName jbiOperation = exchange.getOperation();
            final QName soapAction = SOAPCaller.getSOAPAction(jbiOperation, cdkExtensions);

            // create service client used to invoke WS
            try {
                ServiceClient serviceClient = null;
                try {
                    // The service client options are set during its creation
                    serviceClient = this.soapContext.borrowServiceClient(address, soapAction,
                            exchange.getExchangePattern(), cdkExtensions);

                    // create the in body
                    final OMElement inBodyElement = SOAPCaller.createSOAPBodyContent(in, exchange
                            .getOperation(), serviceClient, cdkExtensions);

                    if (this.logger.isLoggable(Level.FINE)) {
                        // The condition is splitted for performance reasons
                        if (inBodyElement != null) {
                            this.logger.log(Level.FINE, "OUTGOING Payload : "
                                    + inBodyElement.toString());
                        }
                    }

                    if (exchange.isInOnlyPattern()) {
                        // send as InOnly message
                        serviceClient.fireAndForget(soapAction, inBodyElement);

                    } else if (exchange.isRobustInOnlyPattern()) {
                        // send as RobustInOnly message
                        // If the robust in only exchange throws a Fault
                        // It will be catch by the component-common framework
                        serviceClient.sendRobust(soapAction, inBodyElement);

                    } else if (exchange.isInOptionalOutPattern() || exchange.isInOutPattern()) {
                        // Performance notification creation
                        PerformanceNotification performanceNotification = null;
                        performanceNotification = this.component
                                .createPerformanceNotificationSendingToOutside(address, exchange);

                        // send as In(optional)Out message
                        SOAPBody outBodyElement = null;
                        try {
                            outBodyElement = serviceClient.sendReceiveBody(soapAction,
                                    inBodyElement, getHeaders(in, cdkExtensions));
                        } catch (final AxisFault e) {
                            // only catched to send the performance
                            // notification.
                            // TODO : Extend to all MEPs
                            throw e;
                        } finally {
                            // Performance notification sending
                            this.component.sendPerformanceNotification(performanceNotification);
                        }

                        // if msg exchange required a response, set it
                        if (outBodyElement != null) {
                            // For performance reasons, the AXIOM's cache is not
                            // used to optimize the namespace writing. The
                            // service
                            // provider must be in charge of this optimization.
                            if (this.logger.isLoggable(Level.FINE)) {
                                this.logger.log(Level.FINE, "RESPONSE Payload : "
                                        + outBodyElement.toString());
                            }

                            OMElement response = buildResponseElement(outBodyElement);
                            out.setContent(AxiomUtils.createSource(response));
                            Marshaller.copyAttachments(response, out);
                        } else {
                            this.logger.log(Level.FINE, "RESPONSE Payload : No response.");
                        }
                    } else {
                        throw new MessagingException("WSCaller - MEP not recognized : "
                                + exchange.getPattern().toString());
                    }
                } finally {
                    this.soapContext.returnServiceClient(address, soapAction, exchange
                            .getExchangePattern(), serviceClient);
                }
            } catch (final Exception e) {
                this.logger.warning("Catch an exception on the WS invocation : " + e.getMessage());

                // The exception is an AxisFault (probably a SOAP Fault)
                if (e instanceof AxisFault) {
                    this.handleSOAPFault(exchange, (AxisFault) e);
                } else {
                    this.handleException(exchange, e);
                }
            }
        }
    }

    /**
     * Retrieve data from JBI properties wich will be set into the SOAP header.
     * 
     * @param nm
     * @return
     */
    private Map<String, DocumentFragment> getHeaders(NormalizedMessage nm,
            ConfigurationExtensions cdkExtensions) {
        Map<String, DocumentFragment> result = new HashMap<String, DocumentFragment>();
        List<String> filters = SUPropertiesHelper.retrieveHeaderList(cdkExtensions);

        // 1. get the properties defined in the SU from the filter value
        if (filters != null && filters.size() > 0) {
            Set<String> properties = new HashSet<String>();
            Set names = nm.getPropertyNames();
            for (Object object : names) {
                if (object instanceof String) {
                    String propertyName = (String) object;
                    if (isFilteredValue(propertyName, filters)) {
                        properties.add(propertyName);
                    }
                }
            }

            for (String propertyName : properties) {
                Object property = nm.getProperty(propertyName);
                // TODO : Inject not only document fragments...
                if (property instanceof DocumentFragment) {
                    result.put(propertyName, (DocumentFragment) property);
                }
            }
        }

        // 2. additional headers
        if (SUPropertiesHelper.retrieveInjectHeader(cdkExtensions)) {
            // add the protocol headers (the ones injected in consumer mode)
            final Object protocolHeadersPropertyObject = nm.getProperty(PROTOCOL_HEADERS);
            if (protocolHeadersPropertyObject != null
                    && protocolHeadersPropertyObject instanceof Map) {
                result.putAll((Map<String, DocumentFragment>) protocolHeadersPropertyObject);
            }
        }
        return result;
    }

    /**
     * 
     * @param propertyName
     * @param filters
     * @return
     */
    protected boolean isFilteredValue(String propertyName, List<String> filters) {
        boolean result = false;
        for (String filter : filters) {
            if (propertyName.equalsIgnoreCase(filter)) {
                return true;
            }

            if (filter.endsWith("*")) {
                String tmp = filter.substring(0, filter.lastIndexOf("*"));
                if (propertyName.equals(tmp) || propertyName.startsWith(tmp)) {
                    return true;
                }
            }
        }
        return result;
    }

    /**
     * 
     * @param body
     * @return
     */
    protected OMElement buildResponseElement(SOAPBody body) {
        OMElement outBodyElement = body.getFirstElement();

        java.util.Iterator<OMElement> it = body.getChildElements();
        OMElement message = null;
        OMElement firstRef = null;
        if (it.hasNext()) {
            message = it.next();
        }
        if (it.hasNext()) {
            firstRef = it.next();
        }

        // create new body without references
        if (firstRef != null) {
            SOAP11Factory factory = new SOAP11Factory();
            OMElement newBodyElement = factory.createOMElement(message.getQName());

            // TODO: find and replace all references: used recursivity
            if (message.getFirstElement() != null) {
                OMElement newParamElement = factory.createOMElement(message.getFirstElement()
                        .getQName());
                Iterator<OMElement> itChildRef = firstRef.getChildElements();
                OMElement refElmt = null;
                while (itChildRef.hasNext()) {
                    refElmt = itChildRef.next();

                    if ((refElmt.getLocalName().equals("multiRef") && (refElmt
                            .getAttribute(new QName("http://schemas.xmlsoap.org/soap/encoding/",
                                    "arrayType")) != null))) {
                        refElmt.setLocalName("array");
                    }

                    newParamElement.addChild(refElmt);
                }

                newBodyElement.addChild(newParamElement);

                outBodyElement = newBodyElement;
            }
        }

        return outBodyElement;
    }

    /**
     * Set the Exception in the JBI Message as Fault
     * 
     * @param exchange
     * @param e
     */
    private void handleException(final Exchange exchange, final Exception e) {
        exchange.setError(new MessagingException("BC-SOAP Exception => " + e.getMessage(), e));
    }

    /**
     * Handle a SOAP Fault. Set it in the JBI message as Fault.
     * 
     * @param exchange
     * @param fault
     */
    private void handleSOAPFault(final Exchange exchange, final AxisFault fault) {
        if ((fault.getFaultMessageContext() != null)
                && (fault.getFaultMessageContext().getEnvelope() != null)) {
            try {
                final Source sourceFault = Marshaller.createSource(fault.getFaultMessageContext()
                        .getEnvelope(), false);
                final Fault jbiFault = exchange.createFault();
                jbiFault.setContent(sourceFault);
                exchange.setFault(jbiFault);
            } catch (final Exception e1) {
                this.logger.log(Level.SEVERE, "Can't return fault to consumer", e1);
            }
        } else {
            this.handleException(exchange, fault);
        }
    }

    /**
     * Create the SOAPBody content. The JBI content is used as root element. All
     * the JBI attachments are added as root element children.
     * 
     * @param body
     * @return the soap body content or null if {@link NormalizedMessage} is
     *         empty.
     * @throws XMLStreamException
     */
    @SuppressWarnings("unchecked")
    protected static OMElement createSOAPBodyContent(final NormalizedMessage nm,
            final QName operation, final ServiceClient client,
            final ConfigurationExtensions extensions) throws XMLStreamException {

        OMElement document = null;
        final Source src = nm.getContent();

        if (src != null) {
            // create the root element
            final OMFactory fac = OMAbstractFactory.getOMFactory();
            document = SOAPCaller.buildOMElementFromSource(src, operation, extensions, fac);

            // add attachments if any
            if (nm.getAttachmentNames().size() > 0) {
                // enable MTOM
                client.getOptions().setProperty(Constants.Configuration.ENABLE_MTOM,
                        Constants.VALUE_TRUE);

                // Add JBI attachments to the document element
                final Set<String> names = nm.getAttachmentNames();
                for (final String key : names) {
                    final DataHandler attachment = nm.getAttachment(key);
                    final OMText attach = fac.createOMText(attachment, true);
                    document.addChild(attach);
                }
            }
        }
        return document;
    }

    /**
     * Get the operation for the output SOAP message. Get the operation defined
     * in the SU first, if null get the JBI operation.
     * 
     * @param jbiOperation
     * @param extensions
     * @return
     */
    protected static QName getSOAPAction(final QName jbiOperation,
            final ConfigurationExtensions extensions) {
        QName soapAction = AbstractExternalServiceCaller.retrieveDefaultSOAPAction(extensions);
        if (soapAction == null) {
            soapAction = jbiOperation;
        }
        return soapAction;
    }

    /**
     * Build the {@link OMElement} from the {@link Source}. The SOAP operation
     * is added as root element if it is configured in the petals extensions.
     * 
     * @param body
     * @param operation
     * @param extensions
     * @return
     */
    protected static OMElement buildOMElementFromSource(final Source body, final QName operation,
            final ConfigurationExtensions extensions, final OMFactory factory)
            throws XMLStreamException {
        final XMLStreamReader parser = StaxUtils.createXMLStreamReader(body);
        final StAXOMBuilder builder = new StAXOMBuilder(parser);
        OMElement rootElement = builder.getDocumentElement();

        if (AbstractExternalServiceCaller.retrieveAddRoot(extensions)) {
            final OMElement newRootElement = factory.createOMElement(operation);
            newRootElement.addChild(rootElement);
            rootElement = newRootElement;
        }
        return rootElement;
    }

    public String getCallerType() {
        return "SOAP";
    }
}
